import einops
import torch
from torch import nn


class VitClassTokens(nn.Module):
    def __init__(self, dim: int, num_tokens: int = 1, location="first", init_std=0.02, aggregate="flatten"):
        super().__init__()
        self.dim = dim
        self.location = location
        self.num_tokens = num_tokens
        self.init_std = init_std
        self.aggregate = aggregate
        if num_tokens > 0:
            if location in ["first", "middle", "last"]:
                self.tokens = nn.Parameter(torch.zeros(1, num_tokens, dim))
            elif location == "bilateral":
                assert num_tokens % 2 == 0
                self.tokens = nn.Parameter(torch.zeros(1, num_tokens, dim))
            elif location == "uniform":
                self.tokens = nn.Parameter(torch.zeros(1, num_tokens, dim))
            else:
                raise NotImplementedError
        else:
            self.tokens = None
        self.reset_parameters()

    def reset_parameters(self):
        if self.num_tokens > 0:
            nn.init.normal_(self.tokens, std=self.init_std)

    def forward(self, x):
        if self.num_tokens == 0:
            return x
        assert x.ndim == 3
        if self.location == "first":
            tokens = self.tokens.expand(len(x), -1, -1)
            x = torch.concat([tokens, x], dim=1)
        elif self.location == "middle":
            tokens = self.tokens.expand(len(x), -1, -1)
            pre, post = x.chunk(chunks=2, dim=1)
            x = torch.concat([pre, tokens, post], dim=1)
        elif self.location == "last":
            tokens = self.tokens.expand(len(x), -1, -1)
            x = torch.concat([x, tokens], dim=1)
        elif self.location == "bilateral":
            first, last = self.tokens.chunk(chunks=2, dim=1)
            first = first.expand(len(x), -1, -1)
            last = last.expand(len(x), -1, -1)
            x = torch.concat([first, x, last], dim=1)
        elif self.location == "uniform":
            chunks = x.chunk(chunks=self.num_tokens + 1, dim=1)
            tokens = self.tokens.expand(len(x), -1, -1)
            # interweave chunk with token
            interweaved = [chunks[0]]
            for i in range(self.num_tokens):
                interweaved.append(tokens[:, i:i + 1])
                interweaved.append(chunks[i + 1])
            x = torch.concat(interweaved, dim=1)
        else:
            raise NotImplementedError
        return x

    @property
    def output_shape(self):
        if self.aggregate == "flatten":
            return self.dim * self.num_tokens,
        if self.aggregate == "mean":
            return self.dim,
        raise NotImplementedError

    def split(self, x):
        if self.num_tokens == 0:
            return None, x
        assert x.ndim == 3
        if self.location == "first":
            cls_tokens = x[:, :self.num_tokens]
            patch_tokens = x[:, self.num_tokens:]
        else:
            raise NotImplementedError
        return cls_tokens, patch_tokens

    def pool(self, x):
        if self.num_tokens == 0:
            raise NotImplementedError
        else:
            # extract tokens
            if self.location == "first":
                x = x[:, :self.num_tokens]
            elif self.location == "middle":
                middle = x.size(1) // 2
                half_num_tokens = self.num_tokens // 2
                start = middle - half_num_tokens
                end = start + self.num_tokens
                x = x[:, start:end]
            elif self.location == "last":
                x = x[:, -self.num_tokens:]
            elif self.location == "bilateral":
                num_tokens_half = self.num_tokens // 2
                x = torch.concat([x[:, :num_tokens_half], x[:, -num_tokens_half:]], dim=1)
            elif self.location == "uniform":
                # all but the last chunk are full
                chunk_size = (x.size(1) - self.num_tokens) // (self.num_tokens + 1) + 1
                x = torch.stack([x[:, (i + 1) * chunk_size + i] for i in range(self.num_tokens)], dim=1)
            else:
                raise NotImplementedError

            # aggregate if multiple tokens are used
            if self.aggregate == "flatten":
                return x.flatten(start_dim=1)
            elif self.aggregate == "mean":
                return x.mean(dim=1)
            else:
                raise NotImplementedError



import collections.abc
from itertools import repeat


def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            assert len(x) == n
            return x
        return tuple(repeat(x, n))

    return parse


def _is_ntuple(n):
    def check(x):
        return isinstance(x, tuple) and len(param) == n

    return check


def to_ntuple(x, n):
    return _ntuple(n=n)(x)


def is_ntuple(x, n):
    return _is_ntuple(n=n)(x)


to_2tuple = _ntuple(2)
is_2tuple = _is_ntuple(2)



from functools import partial

from torch import nn

from kappamodules.attention import MMDiTDotProductAttention
from kappamodules.modulation import Dit
from kappamodules.modulation.functional import modulate_scale_shift, modulate_gate
from .mlp import Mlp


class MMDitBlock(nn.Module):
    """ multi-modal adaptive norm block (https://arxiv.org/abs/2403.03206) """

    def __init__(
            self,
            dim,
            num_heads,
            mlp_hidden_dim=None,
            cond_dim=None,
            qkv_bias=True,
            attn_ctor=MMDiTDotProductAttention,
            act_ctor=nn.GELU,
            eps=1e-6,
            init_weights="xavier_uniform",
            init_last_proj_zero=False,
            init_gate_zero=False,
    ):
        super().__init__()
        # DiT uses non-affine LayerNorm and GELU with tanh-approximation
        norm_ctor = partial(nn.LayerNorm, elementwise_affine=False)
        # properties
        mlp_hidden_dim = mlp_hidden_dim or dim * 4
        cond_dim = cond_dim or dim
        # modulation
        self.modulation = Dit(
            cond_dim=cond_dim,
            out_dim=dim,
            init_weights=init_weights,
            num_outputs=12,
            gate_indices=[2, 5, 8, 11],
            init_gate_zero=init_gate_zero,
        )
        # attn
        self.norm1 = norm_ctor(dim, eps=eps)
        self.attn = attn_ctor(
            dim=dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        # mlp
        self.norm2 = norm_ctor(dim, eps=eps)
        self.mlp1 = Mlp(
            in_dim=dim,
            hidden_dim=mlp_hidden_dim,
            act_ctor=act_ctor,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.mlp2 = Mlp(
            in_dim=dim,
            hidden_dim=mlp_hidden_dim,
            act_ctor=act_ctor,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )

    def _attn_residual_path(self, x1, x2, scale1, shift1, gate1, scale2, shift2, gate2, attn_mask):
        x1 = modulate_scale_shift(self.norm1(x1), scale=scale1, shift=shift1)
        x2 = modulate_scale_shift(self.norm1(x2), scale=scale2, shift=shift2)
        x1, x2 = self.attn(x1=x1, x2=x2, attn_mask=attn_mask)
        x1 = modulate_gate(x1, gate=gate1)
        x2 = modulate_gate(x2, gate=gate2)
        return x1, x2

    def _mlp_residual_path(self, x1, x2, scale1, shift1, gate1, scale2, shift2, gate2):
        x1 = modulate_gate(self.mlp1(modulate_scale_shift(self.norm2(x1), scale=scale1, shift=shift1)), gate=gate1)
        x2 = modulate_gate(self.mlp2(modulate_scale_shift(self.norm2(x2), scale=scale2, shift=shift2)), gate=gate2)
        return x1, x2

    def forward(self, x1, x2, cond, attn_mask=None):
        scale1, shift1, g1, scale2, shift2, g2, scale3, shift3, g3, scale4, shift4, g4 = self.modulation(cond)
        x1, x2 = self._attn_residual_path(
            x1=x1,
            x2=x2,
            scale1=scale1,
            shift1=shift1,
            gate1=g1,
            scale2=scale2,
            shift2=shift2,
            gate2=g2,
            attn_mask=attn_mask,
        )
        x1, x2 = self._mlp_residual_path(
            x1=x1,
            x2=x2,
            scale1=scale3,
            shift1=shift3,
            gate1=g3,
            scale2=scale4,
            shift2=shift4,
            gate2=g4,
        )
        return x1, x2



import torch.nn.functional as F
from torch import nn


class L1Loss(nn.Module):
    @staticmethod
    def forward(pred, target, reduction="mean"):
        return F.l1_loss(pred, target, reduction=reduction)



def modulate_scale_shift(x, scale, shift):
    if x.ndim == 3:
        scale = scale.unsqueeze(1)
        shift = shift.unsqueeze(1)
    return x * (1 + scale) + shift


def modulate_gate(x, gate):
    if x.ndim == 3:
        gate = gate.unsqueeze(1)
    return gate * x



import einops
import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import nn


class DataNormStateDictPreHook:
    def __call__(self, module, *args, **kwargs):
        if dist.is_initialized():
            module.finish()


class DataNorm(nn.Module):
    def __init__(self, dim, eps=1e-6, channel_first=True):
        super().__init__()
        self.dim = dim
        self.eps = eps
        self.channel_first = channel_first
        self.register_buffer("mean", torch.zeros(dim))
        self.register_buffer("var", torch.ones(dim))
        self.register_buffer("num_batches_tracked", torch.tensor(0.))
        self.mean_buffer = []
        self.var_buffer = []
        self._async_handle = None
        self.register_state_dict_pre_hook(DataNormStateDictPreHook())

    def _x_to_stats(self, x):
        # handles all use-cases
        # channel_first
        # - x.ndim == 2 (dim,) -> average over dim=[0]
        # - x.ndim == 3 (dim, height) -> average over dim=[0, 2]
        # - x.ndim == 4 (dim, height, width) -> average over dim=[0, 2, 3]
        # - x.ndim == 5 (dim, height, width, depth) -> average over dim=[0, 2, 3, 4]
        # channel_last
        # - x.ndim == 2 (dim,) -> average over dim=[0]
        # - x.ndim == 3 (height, dim) -> average over dim=[0, 1]
        # - x.ndim == 4 (height, width, dim) -> average over dim=[0, 1, 2]
        # - x.ndim == 5 (height, width, depth, dim) -> average over dim=[0, 1, 2, 3]
        if self.channel_first:
            x = einops.rearrange(x, "bs dim ... -> (bs ...) dim")
        else:
            x = einops.rearrange(x, "bs ... dim -> (bs ...) dim")
        mean = x.mean(dim=0)
        var = x.var(dim=0, unbiased=False)
        return mean, var

    def finish(self):
        self._update_stats(inplace=True)

    def _update_stats(self, inplace):
        if self._async_handle is not None:
            # get from async handle
            self._async_handle.wait()
            x = self._async_handle.get_future().value()
            x = torch.concat(x)
            self._async_handle = None
            # add stats to buffer
            xmean, xvar = self._x_to_stats(x)
            self.mean_buffer.append(xmean)
            self.var_buffer.append(xvar)

        # update should be called on every update
        assert len(self.mean_buffer) == 1
        assert len(self.var_buffer) == 1

        # accumulate stats
        mean = self.mean_buffer[0]
        var = self.var_buffer[0]
        # clear buffers
        self.mean_buffer.clear()
        self.var_buffer.clear()
        # calculate momentum
        self.num_batches_tracked += 1
        momentum = 1 / self.num_batches_tracked
        # update stats
        if inplace:
            # if used in nograd environment -> inplace
            self.mean.mul_(1 - momentum).add_(mean, alpha=momentum)
            self.var.mul_(1 - momentum).add_(var, alpha=momentum)
        else:
            # if used in grad environment -> old mean/var are required for backward
            self.mean = (self.mean * (1 - momentum)).add_(mean, alpha=momentum)
            self.var = (self.var * (1 - momentum)).add_(var, alpha=momentum)

    def forward(self, x):
        if self.training:
            if len(x) == 1:
                raise NotImplementedError("DataNorm batch_size=1 requires syncing features instead of stats")

        # multi GPU -> queue communication of batch stats
        if dist.is_initialized():
            # update stats for previous iteration
            if self._async_handle is not None:
                self._update_stats(inplace=True)
            # queue synchonization for current iteration
            if self.training:
                assert self._async_handle is None
                tensor_list = [torch.zeros_like(x) for _ in range(dist.get_world_size())]
                self._async_handle = dist.all_gather(tensor_list, x, async_op=True)

        # normalize
        og_x = x
        if not self.channel_first:
            x = einops.rearrange(x, "bs ... dim -> bs dim ...")
        x = F.batch_norm(
            input=x,
            running_mean=self.mean,
            running_var=self.var,
            eps=self.eps,
            # avoid updating mean/var
            training=False,
        )
        if not self.channel_first:
            x = einops.rearrange(x, "bs dim ... -> bs ... dim")

        # single GPU -> directly update stats
        if self.training and not dist.is_initialized():
            with torch.no_grad():
                xmean, xvar = self._x_to_stats(og_x)
                self.mean_buffer.append(xmean)
                self.var_buffer.append(xvar)
            self._update_stats(inplace=not x.requires_grad)

        return x



# adapted from https://github.com/microsoft/pdearena/blob/main/pdearena/modules/conditioned/twod_unet.py
import torch
from torch import nn


class ResidualBlock(nn.Module):
    """Wide Residual Blocks used in modern Unet architectures.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        cond_channels (int): Number of channels in the conditioning vector.
        norm (bool): Whether to use normalization.
        n_groups (int): Number of groups for group normalization.
        use_scale_shift_norm (bool): Whether to use scale and shift approach to conditoning (also termed as `AdaGN`).
    """

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            cond_channels: int,
            norm: bool = False,
            n_groups: int = 1,
            use_scale_shift_norm: bool = False,
    ):
        super().__init__()
        self.use_scale_shift_norm = use_scale_shift_norm
        self.activation = nn.GELU()

        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=(3, 3), padding=(1, 1))
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=(3, 3), padding=(1, 1))
        # If the number of input channels is not equal to the number of output channels we have to
        # project the shortcut connection
        if in_channels != out_channels:
            self.shortcut = nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1))
        else:
            self.shortcut = nn.Identity()

        if norm:
            self.norm1 = nn.GroupNorm(n_groups, in_channels)
            self.norm2 = nn.GroupNorm(n_groups, out_channels)
        else:
            self.norm1 = nn.Identity()
            self.norm2 = nn.Identity()

        self.cond_emb = nn.Linear(cond_channels, 2 * out_channels if use_scale_shift_norm else out_channels)
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.zeros_(self.conv2.weight)
        torch.nn.init.zeros_(self.conv2.bias)

    def forward(self, x, emb):
        # First convolution layer
        h = self.conv1(self.activation(self.norm1(x)))
        emb_out = self.cond_emb(emb)
        while len(emb_out.shape) < len(h.shape):
            emb_out = emb_out[..., None]
        if self.use_scale_shift_norm:
            scale, shift = torch.chunk(emb_out, 2, dim=1)
            h = self.norm2(h) * (1 + scale) + shift  # where we do -1 or +1 doesn't matter
            h = self.conv2(self.activation(h))
        else:
            h = h + emb_out
            # Second convolution layer
            h = self.conv2(self.activation(self.norm2(h)))
        # Add the shortcut connection and return
        return h + self.shortcut(x)


class AttentionBlock(nn.Module):
    """Attention block This is similar to [transformer multi-head
    attention](https://arxiv.org/abs/1706.03762).

    Args:
        n_channels: the number of channels in the input
        n_heads:  the number of heads in multi-head attention
        d_k: the number of dimensions in each head
        n_groups: the number of groups for [group normalization][torch.nn.GroupNorm]

    """

    def __init__(self, n_channels: int, n_heads: int = 1, d_k=None, n_groups: int = 1):
        """ """
        super().__init__()

        # Default `d_k`
        if d_k is None:
            d_k = n_channels
        # Normalization layer
        self.norm = nn.GroupNorm(n_groups, n_channels)
        # Projections for query, key and values
        self.projection = nn.Linear(n_channels, n_heads * d_k * 3)
        # Linear layer for final transformation
        self.output = nn.Linear(n_heads * d_k, n_channels)
        # Scale for dot-product attention
        self.scale = d_k ** -0.5
        #
        self.n_heads = n_heads
        self.d_k = d_k

    def forward(self, x):
        # Get shape
        batch_size, n_channels, height, width = x.shape
        # Change `x` to shape `[batch_size, seq, n_channels]`
        x = x.view(batch_size, n_channels, -1).permute(0, 2, 1)
        # Get query, key, and values (concatenated) and shape it to `[batch_size, seq, n_heads, 3 * d_k]`
        qkv = self.projection(x).view(batch_size, -1, self.n_heads, 3 * self.d_k)
        # Split query, key, and values. Each of them will have shape `[batch_size, seq, n_heads, d_k]`
        q, k, v = torch.chunk(qkv, 3, dim=-1)
        # Calculate scaled dot-product $\frac{Q K^\top}{\sqrt{d_k}}$
        attn = torch.einsum("bihd,bjhd->bijh", q, k) * self.scale
        # Softmax along the sequence dimension $\underset{seq}{softmax}\Bigg(\frac{Q K^\top}{\sqrt{d_k}}\Bigg)$
        attn = attn.softmax(dim=1)
        # Multiply by values
        res = torch.einsum("bijh,bjhd->bihd", attn, v)
        # Reshape to `[batch_size, seq, n_heads * d_k]`
        res = res.view(batch_size, -1, self.n_heads * self.d_k)
        # Transform to `[batch_size, seq, n_channels]`
        res = self.output(res)

        # Add skip connection
        res += x

        # Change to shape `[batch_size, in_channels, height, width]`
        res = res.permute(0, 2, 1).view(batch_size, n_channels, height, width)
        return res


class DownBlock(nn.Module):
    """Down block This combines `ResidualBlock` and `AttentionBlock`.

    These are used in the first half of U-Net at each resolution.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        cond_channels (int): Number of channels in the conditioning vector.
        has_attn (bool): Whether to use attention block
        norm (bool): Whether to use normalization
        use_scale_shift_norm (bool): Whether to use scale and shift approach to conditoning (also termed as `AdaGN`).
    """

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            cond_channels: int,
            has_attn: bool = False,
            norm: bool = False,
            use_scale_shift_norm: bool = False,
    ):
        super().__init__()
        self.res = ResidualBlock(
            in_channels,
            out_channels,
            cond_channels,
            norm=norm,
            use_scale_shift_norm=use_scale_shift_norm,
        )
        if has_attn:
            self.attn = AttentionBlock(out_channels)
        else:
            self.attn = nn.Identity()

    def forward(self, x, emb):
        x = self.res(x, emb)
        x = self.attn(x)
        return x


class UpBlock(nn.Module):
    """Up block This combines `ResidualBlock` and `AttentionBlock`.

    These are used in the second half of U-Net at each resolution.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        cond_channels (int): Number of channels in the conditioning vector.
        has_attn (bool): Whether to use attention block
        norm (bool): Whether to use normalization
        use_scale_shift_norm (bool): Whether to use scale and shift approach to conditoning (also termed as `AdaGN`).
    """

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            cond_channels: int,
            has_attn: bool = False,
            norm: bool = False,
            use_scale_shift_norm: bool = False,
    ):
        super().__init__()
        # The input has `in_channels + out_channels` because we concatenate the output of the same resolution
        # from the first half of the U-Net
        self.res = ResidualBlock(
            in_channels + out_channels,
            out_channels,
            cond_channels,
            norm=norm,
            use_scale_shift_norm=use_scale_shift_norm,
        )
        if has_attn:
            self.attn = AttentionBlock(out_channels)
        else:
            self.attn = nn.Identity()

    def forward(self, x, emb) -> torch.Tensor:
        x = self.res(x, emb)
        x = self.attn(x)
        return x


class MiddleBlock(nn.Module):
    """Middle block It combines a `ResidualBlock`, `AttentionBlock`, followed by another
    `ResidualBlock`.

    This block is applied at the lowest resolution of the U-Net.

    Args:
        n_channels (int): Number of channels in the input and output.
        cond_channels (int): Number of channels in the conditioning vector.
        has_attn (bool, optional): Whether to use attention block. Defaults to False.
        norm (bool, optional): Whether to use normalization. Defaults to False.
        use_scale_shift_norm (bool, optional): Whether to use scale and shift approach to conditoning (also termed as `AdaGN`). Defaults to False.
    """

    def __init__(
            self,
            n_channels: int,
            cond_channels: int,
            has_attn: bool = False,
            norm: bool = False,
            use_scale_shift_norm: bool = False,
    ):
        super().__init__()
        self.res1 = ResidualBlock(
            n_channels,
            n_channels,
            cond_channels,
            norm=norm,
            use_scale_shift_norm=use_scale_shift_norm,
        )
        self.attn = AttentionBlock(n_channels) if has_attn else nn.Identity()
        self.res2 = ResidualBlock(
            n_channels,
            n_channels,
            cond_channels,
            norm=norm,
            use_scale_shift_norm=use_scale_shift_norm,
        )

    def forward(self, x, emb) -> torch.Tensor:
        x = self.res1(x, emb)
        x = self.attn(x)
        x = self.res2(x, emb)
        return x


class Upsample(nn.Module):
    r"""Scale up the feature map by $2 \times$"""

    def __init__(self, n_channels: int):
        super().__init__()
        self.conv = nn.ConvTranspose2d(n_channels, n_channels, (4, 4), (2, 2), (1, 1))

    def forward(self, x):
        return self.conv(x)


class Downsample(nn.Module):
    r"""Scale down the feature map by $\frac{1}{2} \times$"""

    def __init__(self, n_channels):
        super().__init__()
        self.conv = nn.Conv2d(n_channels, n_channels, (3, 3), (2, 2), (1, 1))

    def forward(self, x):
        return self.conv(x)


class UnetPdearena(nn.Module):
    """Modern U-Net architecture

    This is a modern U-Net architecture with wide-residual blocks and spatial attention blocks

    Args:
        input_dim (int): Number input dimensions
        output_dim (int): Number of output dimensions
        hidden_channels (int): Number of channels in the hidden layers
        norm (bool): Whether to use normalization
        ch_mults (list): List of channel multipliers for each resolution
        is_attn (list): List of booleans indicating whether to use attention blocks
        mid_attn (bool): Whether to use attention block in the middle block
        n_blocks (int): Number of residual blocks in each resolution
        param_conditioning (Optional[str]): Type of conditioning to use. Defaults to None.
        use_scale_shift_norm (bool): Whether to use scale and shift approach to conditoning (also termed as `AdaGN`). Defaults to False.
        use1x1 (bool): Whether to use 1x1 convolutions in the initial and final layers

    Note:
        Currently, only `scalar` parameter conditioning is supported.
    """

    def __init__(
            self,
            input_dim: int,
            output_dim: int,
            hidden_channels,
            cond_dim: int = None,
            norm: bool = False,
            ch_mults=(1, 2, 2, 4),
            is_attn=(False, False, False, False),
            mid_attn: bool = False,
            n_blocks: int = 2,
            use_scale_shift_norm: bool = False,
            use1x1: bool = False,
    ) -> None:
        super().__init__()
        self.hidden_channels = hidden_channels
        self.activation = nn.GELU()

        # Number of resolutions
        n_resolutions = len(ch_mults)

        n_channels = hidden_channels

        # Project image into feature map
        if use1x1:
            self.image_proj = nn.Conv2d(input_dim, n_channels, kernel_size=1)
        else:
            self.image_proj = nn.Conv2d(input_dim, n_channels, kernel_size=(3, 3), padding=(1, 1))

        # #### First half of U-Net - decreasing resolution
        down = []
        # Number of channels
        out_channels = in_channels = n_channels
        # For each resolution
        for i in range(n_resolutions):
            # Number of output channels at this resolution
            out_channels = in_channels * ch_mults[i]
            # Add `n_blocks`
            for _ in range(n_blocks):
                down.append(
                    DownBlock(
                        in_channels,
                        out_channels,
                        cond_dim,
                        has_attn=is_attn[i],
                        norm=norm,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                )
                in_channels = out_channels
            # Down sample at all resolutions except the last
            if i < n_resolutions - 1:
                down.append(Downsample(in_channels))

        # Combine the set of modules
        self.down = nn.ModuleList(down)

        # Middle block
        self.middle = MiddleBlock(
            out_channels,
            cond_dim,
            has_attn=mid_attn,
            norm=norm,
            use_scale_shift_norm=use_scale_shift_norm,
        )

        # #### Second half of U-Net - increasing resolution
        up = []
        # Number of channels
        in_channels = out_channels
        # For each resolution
        for i in reversed(range(n_resolutions)):
            # `n_blocks` at the same resolution
            out_channels = in_channels
            for _ in range(n_blocks):
                up.append(
                    UpBlock(
                        in_channels,
                        out_channels,
                        cond_dim,
                        has_attn=is_attn[i],
                        norm=norm,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                )
            # Final block to reduce the number of channels
            out_channels = in_channels // ch_mults[i]
            up.append(
                UpBlock(
                    in_channels,
                    out_channels,
                    cond_dim,
                    has_attn=is_attn[i],
                    norm=norm,
                    use_scale_shift_norm=use_scale_shift_norm,
                )
            )
            in_channels = out_channels
            # Up sample at all resolutions except last
            if i > 0:
                up.append(Upsample(in_channels))

        # Combine the set of modules
        self.up = nn.ModuleList(up)

        if norm:
            self.norm = nn.GroupNorm(8, n_channels)
        else:
            self.norm = nn.Identity()
        if use1x1:
            self.final = nn.Conv2d(in_channels, output_dim, kernel_size=1)
        else:
            self.final = nn.Conv2d(in_channels, output_dim, kernel_size=(3, 3), padding=(1, 1))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.zeros_(self.final.weight)
        torch.nn.init.zeros_(self.final.bias)

    def forward(self, x, emb):
        assert x.ndim == 4

        x = self.image_proj(x)

        h = [x]
        for m in self.down:
            if isinstance(m, Downsample):
                x = m(x)
            else:
                x = m(x, emb)
            h.append(x)

        x = self.middle(x, emb)

        for m in self.up:
            if isinstance(m, Upsample):
                x = m(x)
            else:
                # Get the skip connection from first half of U-Net and concatenate
                s = h.pop()
                x = torch.cat((x, s), dim=1)
                #
                x = m(x, emb)

        x = self.final(self.activation(self.norm(x)))
        return x



import einops
import torch
import torch.nn.functional as F
from torch import nn

from kappamodules.layers import LinearProjection


class CrossAttentionPooling(nn.Module):
    def __init__(
            self,
            dim: int,
            num_heads: int = 12,
            out_dim: int = None,
            kv_bias: bool = True,
            num_query_tokens: int = 1,
            init_std: float = 0.02,
            norm_ctor=None,
            init_weights: str = "truncnormal002",
    ):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.dim = dim
        out_dim = out_dim or dim
        self.out_dim = out_dim
        self.head_dim = dim // num_heads
        self.num_heads = num_heads
        self.init_std = init_std

        self.norm = nn.Identity() if norm_ctor is None else norm_ctor(dim)

        if num_query_tokens > 0:
            # transformer doesnt have a CLS token -> learn it in the pooling
            self.query_tokens = nn.Parameter(torch.zeros(1, num_query_tokens, dim))
        else:
            # use CLS token of transformer
            self.query_tokens = None
        self.kv = LinearProjection(dim, dim * 2, bias=kv_bias, init_weights=init_weights)
        self.out = LinearProjection(dim, out_dim, init_weights=init_weights, optional=True)

        self.num_query_tokens = num_query_tokens
        self.reset_parameters()

    def reset_parameters(self):
        if self.query_tokens is not None:
            nn.init.normal_(self.query_tokens, std=self.init_std)
        self.kv.reset_parameters()

    def forward(self, x, query_tokens=None):
        if self.query_tokens is None:
            assert query_tokens is not None, f"num_query_tokens == 0 -> requires external query tokens"
            # check of provided tokens is (batch_size, num_query_tokens, dim)
            assert query_tokens.ndim == 3
            assert len(query_tokens) == len(x)
            assert query_tokens.size(2) == self.dim
        else:
            assert query_tokens is None, f"attention pooling learns own query tokens"

        # create query
        if query_tokens is None:
            query_tokens = self.query_tokens.expand(len(x), -1, -1)
        q = einops.rearrange(
            query_tokens,
            "bs seqlen_q (num_heads head_dim) -> bs num_heads seqlen_q head_dim",
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        )

        # create kv
        kv = self.kv(self.norm(x))
        k, v = einops.rearrange(
            kv,
            "bs seqlen_kv (two num_heads head_dim) -> two bs num_heads seqlen_kv head_dim",
            two=2,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)

        x = F.scaled_dot_product_attention(q, k, v)
        x = einops.rearrange(x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        x = self.out(x)

        return x



from torch import nn


class IgnoreArgsAndKwargsWrapper(nn.Module):
    """
    wrapper for any nn.Module that ignores everything besides the first arg passed to forward
    use-case: pytorch_geometrics InstanceNorm
    self.message = km.Sequential(
      km.IgnoreArgsAndKwargsWrapper(nn.Linear(10, 10)),
      pyg.InstanceNorm(10),
      ...
    )
    y = self.message(x, batch=batch)
    """
    def __init__(self, module):
        super().__init__()
        self.module = module

    def forward(self, x, *_, **__):
        return self.module(x)


from torch import nn

from kappamodules.init import init_xavier_uniform_zero_bias, init_truncnormal_zero_bias, init_norms_as_noaffine
from kappamodules.utils.mode_to_ctor import mode_to_norm_ctor


class ContrastiveProjector(nn.Module):
    def __init__(
            self,
            input_dim,
            hidden_dim=2048,
            num_hidden_layers=1,
            output_dim=256,
            norm_mode="batchnorm",
            init_weights="xavier_uniform",
            eps=1e-6,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_hidden_layers = num_hidden_layers
        self.output_dim = output_dim
        self.norm_mode = norm_mode
        self.init_weights = init_weights
        self.eps = eps

        # ctors
        self.norm_ctor, self.requires_bias = mode_to_norm_ctor(norm_mode)
        self.act_ctor = nn.GELU

        # initial layer
        self.first_layer = nn.Sequential(
            nn.Linear(input_dim, hidden_dim, bias=self.requires_bias),
            self.norm_ctor(hidden_dim, eps=eps),
            self.act_ctor(),
        )
        # hidden layers
        self.hidden_layers = nn.Sequential(
            *[
                nn.Sequential(
                    nn.Linear(hidden_dim, hidden_dim, bias=self.requires_bias),
                    self.norm_ctor(hidden_dim, eps=eps),
                    self.act_ctor(),
                )
                for _ in range(num_hidden_layers)
            ]
        )
        # last layer
        self.last_layer = nn.Sequential(
            nn.Linear(hidden_dim, output_dim, bias=self.requires_bias),
            self.norm_ctor(hidden_dim, eps=eps),
        )
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        self.apply(init_norms_as_noaffine)

    def forward(self, x):
        x = self.first_layer(x)
        x = self.hidden_layers(x)
        x = self.last_layer(x)
        return x
